PACKAGE_PARENT = '.'

try:
    from . import ratelimit
    from .cfg import AUTH_HEADER as HEADER
    from .cfg import API_URL as URL
except (ImportError, ModuleNotFoundError):
    import ratelimit
    from cfg import AUTH_HEADER as HEADER
    from cfg import API_URL as URL


class User:
    """Represents a discord member. Used internally to parse interaction/member JSON data.

    :param json memberJson: JSON to parse into this class.

    .. Important::
        TODO: support sernding dms to user - add a method to send a DM to the user on the user object

        TODO: Support avatar urls for Users - add an `avatar_url` property in the User class"""

    def __init__(self, memberJson):
        self.memberJson = memberJson

    @property
    def avatar(self):
        """
        :return: the member's avatar hash
        :rtype: str
        .. Important::
            TODO:

            - make it an ``avatar_url`` or make that a separate property

        """
        return self.memberJson['avatar']

    @property
    def id(self):
        """
        :return: the user's id
        :rtype: int
        """
        return int(self.memberJson['id'])

    @property
    def username(self):
        """
        :return: the user's username
        :rtype: str
        """
        return self.memberJson['username']

    @property
    def discriminator(self):
        """
        :return: the user's discriminator
        :rtype: int
        """

        return int(self.memberJson['discriminator'])


class Member:
    """
    Discord Guild Member Object.

    Contains the discordSplash.member.User object but this class has additional support for roles and other guild specific data.

    .. SeeAlso::
        - Also known as a Guild Member
        - discordSplash.member.User
    """

    def __init__(self, json):
        self.json = json

    @property
    def user(self):
        """
        Discord user that is in the guild

        :return: discordSplash.member.User object
        :rtype: discordSplash.member.User
        """
        try:
            return User(self.json['user'])
        except KeyError:
            return None

    @property
    def nickname(self):
        """
        Nickname of the user

        :return: Member's nickname
        :rtype: str
        """
        try:
            return self.json['user']
        except KeyError:
            return None

    @property
    def roles(self):
        """
        The Member's Roles

        .. Important::
            TODO:

            - add the Role object

        :return: a list of all of the member's role ids
        :rtype: [int]
        """
        try:
            return self.json['roles']
        except KeyError:
            return None

    @property
    def joined_at(self):
        """
        When the member joined the guild

        :return: ISO8601 Timestamp
        :rtype: str
        """
        try:
            return self.json['joined_at']
        except KeyError:
            return None

    @property
    def guild_id(self):
        """
        ID of the Guild the member is in
        :return: Guild ID
        :rtype: int
        """
        try:
            return int(self.json['guild_id'])
        except KeyError:
            return None

    @property
    def premium_since(self):
        """
        When the member began boosting the guild

        :return: ISO8601 Timestamp
        :rtype: str
        """
        try:
            return self.json['premium_since']
        except KeyError:
            return None

    @property
    def deaf(self):
        """
        Whether or not the user is deafened in the guild

        :return: True if the user is deafened
        :rtype: bool
        """
        try:
            return bool(self.json['deaf'])
        except KeyError:
            return None

    @property
    def mute(self):
        """
        Whether or not the user is muted in the guild

        :return: True if the user is muted
        :rtype: bool
        """
        try:
            return bool(self.json['mute'])
        except KeyError:
            return None

    @property
    def pending(self):
        """
        Whether or not the member has passed the membership screening

        :return: False if the member has passed the screening
        :rtype: bool
        """
        try:
            return bool(self.json['pending'])
        except KeyError:
            return None

    @property
    def permissions(self):
        """
        Total permissions of the member in the channel

        .. Tip::
            This includes channel overrides

        .. Note::
            This only is used during the ``INTERACTION_CREATE`` Gateway event.

        :return:
        :rtype:
        """
        try:
            return self.json['']
        except KeyError:
            return None

    async def update(self, guild_id, nick: str = None, roles: list = None, mute: bool = None, deaf: bool = None,
                     channel_id: int = None):
        """
        modifies the guild member object

        .. Hint::
            All of these fields require permission.

            All of these fields are optional

        .. Warning::
            If ``channel_id`` is not set, the user will be disconnected from voice.

        .. SeeAlso::
            coroutine discordSplash.guild.Guild.modify_member()

        :param str nick: new nickname of the member
        :param [list] roles: list of roles the member has
        :param bool mute: whether or not the user is muted
        :param bool deaf: whether or not the user is deafened
        :param str channel_id: id of the voice channel to move the member.

        :return: updated Member object
        :rtype: discordSplash.member.Member
        """
        json = {"nick": nick, "roles": roles, "mute": mute, "deaf": deaf, "channel_id": channel_id}
        g_id = guild_id
        id_ = self.json['user']['id']

        r = await ratelimit.patch(f'{URL}/guilds/{g_id}/members/{id_}', json=json, guild_id=g_id)
        member_ = Member(r)
        print(r)
        return member_



try:
    from .cfg import AUTH_HEADER as HEADER
    from .cfg import API_URL as URL
    from . import member, channel
except (ImportError, ModuleNotFoundError):
    from cfg import AUTH_HEADER as HEADER
    from cfg import API_URL as URL
    import member
    import channel
import aiohttp

ratelimitJS = {}
# format should look like:
# {
# <guild_id>: {
#       "remaining": 2,
#        "reset": 234877482
#   }
# }


class Guild:
    """Class for a Discord Guild"""

    def __init__(self, jsondata):
        self.json = jsondata

    @property
    def id(self):
        """
        Returns the Guild's id

        :return: id of the guild represented in this object
        :rtype: int
        """
        return int(self.json['id'])

    @property
    def name(self):
        """
        Returns the name of the guild

        :return: name of guild, without leading or trailing whitespace
        :rtype: str
        """
        return self.json['name']

    @property
    def icon(self):
        """
        Icon hash for the guild

        .. Important::
            TODO: Make an ``icon_url`` property - this `icon_url` property will have the actual icon for the guild instead of just the hash.

        :return: Icon hash for the guild
        :rtype: str
        """
        try:
            return self.json['icon']
        except KeyError:
            return None

    @property
    def icon_hash(self):
        """
        Icon hash for the guild if the guild is in a template object

        .. Warning::
            I dont quite understand this. Please open a PR on this

        :return: icon hash of the guild
        :rtype: str
        """
        try:
            return self.json['icon_hash']
        except KeyError:
            return None

    @property
    def invite_splash(self):
        """
        Splash background image for the guild
        Appears when joining a guild from a web browser.

        :return: Image hash for the background image
        :rtype: str
        """
        try:
            return self.json['splash']
        except KeyError:
            return None

    @property
    def discovery_splash(self):
        """
        Discovery Splash hash

        :returns: hash of the Discovery Splash image
        :rtype: str
        """
        try:
            return self.json['discovery_splash']
        except KeyError:
            return None

    @property
    def is_owner(self):
        """
        Is the bot the Guild owner?

        :returns: whether or not the bot is the guild owner
        :rtype: bool
        """
        try:
            return bool(self.json['owner'])
        except KeyError:
            return None

    @property
    def owner_id(self):
        """
        id of the guild owner

        :returns: guild owner's id
        :rtype: int
        """
        try:
            return int(self.json['owner_id'])
        except KeyError:
            return None

    @property
    def permissions(self):
        """
        Permissions for the bot in the guild

        .. Important::
            TODO: make bot guild permissions an ENUM - having this as an enum will make it easier to understand permissions. | eg. ``guild.permissions.BAN_MEMBERS`` instead of something like ``0x39408``


        :return: Total permissions for the current bot in the guild
        :rtype: str
        """
        try:
            return self.json['permissions']
        except KeyError:
            return None

    @property
    def region(self):
        """
        Voice Region id for the guild

        :return: Region id for the guild
        :rtype: str
        """
        try:
            return self.json['region']
        except KeyError:
            return None

    @property
    def afk_channel(self):
        """
        id of the guild's AFK channel

        :return: AFK channel id
        :rtype: int
        """
        try:
            try:
                return int(self.json['afk_channel_id'])
            except ValueError:
                return self.json['afk_channel_id']
        except KeyError:
            return None

    @property
    def afk_timeout(self):
        """
        Time that it takes to be moved into an AFK channel

        :return: Time, in seconds, of the AFK timeout
        :rtype: int
        """
        try:
            return int(self.json['afk_timeout'])
        except KeyError:
            return None

    @property
    def widget_enabled(self):
        """
        Whether or not the server widget is enabled

        :return: ``True`` if the widget is enabled
        :rtype: bool
        """
        try:
            return bool(self.json['widget_enabled'])
        except KeyError:
            return None

    @property
    def widget_channel_id(self):
        """
        Channel ID that the widget invites to. Returns ``None`` if invite is off

        :return: Channel ID of the widget invite
        :rtype: int
        """
        try:
            try:
                return int(self.json['widget_channel_id'])
            except ValueError:
                return None
        except KeyError:
            return None

    @property
    def verification_level(self):
        """
        Verification level required for the guild

        .. Important::
            TODO: Make guild verification_level an ENUM - make the guild verification level an enum | eg. ``guilds.verification_level.phone`` instead of ``4``


        :return: Verification level required to talk in the guild
        :rtype: int
        """
        try:
            return int(self.json['verification_level'])
        except KeyError:
            return None

    @property
    def explicit_content_filter(self):
        """
        Level of content scanning in the guild

        :return: Level of content scanning and auto-deletion in the guild
        :rtype: int
        """
        try:
            return int(self.json['explicit_content_filter'])
        except KeyError:
            return None

    @property
    def default_message_notifications(self):
        """
        Default message notification level

        .. Important::
            TODO: make default_message_notifications an enum -  make default_message_notifications an enum | eg. ``...notifications.all`` instead of ``0``

            - Make it an ENUM

        :return: Default message notification level for the guild
        :rtype: int

        """
        try:
            return int(self.json['default_message_notifications'])
        except KeyError:
            return None

    @property
    def roles(self):
        """
        List of roles found in the guild

        .. Important::
            TODO: add discordSplash.guild.role object - add discordSplash.guild.role object | this would turn the json of the roles into an object.



        .. Danger::
            This does not work.

        :return: list of discordSplash.guild.Role objects
        :rtype: list
        """

        # try:
        #    return self.json['x']
        # except KeyError:
        #    return None
        return print('This has not been implemented yet')

    @property
    def features(self):
        """
        All features found in the guild

        :return: a list of guild feature strings
        :rtype: list
        """
        try:
            return self.json['features']
        except KeyError:
            return None

    @property
    def mfa_level(self):
        """
        Required MFA level for the guild

        :return: MFA level for the guild
        :rtype: int
        """
        try:
            return int(self.json['mfa_level'])
        except KeyError:
            return None

    @property
    def owner_app_id(self):
        """
        If the guild creator is a bot, this returns the bot's application ID.

        :return: Guild owner's app id
        :rtype: int
        """
        try:
            try:
                return int(self.json['application_id'])
            except ValueError:
                return None
        except KeyError:
            return None

    @property
    def system_channel_id(self):
        """
        The Channel ID for system messages (member join, boosts).

        :return: System messages channel id
        :rtype: int
        """
        try:
            try:
                return int(self.json['system_channel_id'])
            except ValueError:
                return None
        except KeyError:
            return None

    @property
    def system_channel_flags(self):
        """
        Flags for the System Channel

        :return: System channel's flags
        :rtype: int
        """
        try:
            return int(self.json['system_channel_flags'])
        except KeyError:
            return None

    @property
    def rules_channel(self):
        """
        Channel where community servers post rules

        :return: Channel id of the rules channel
        :rtype: int
        """
        try:
            return int(self.json['x'])
        except KeyError:
            return None

    async def get_preview(self):
        """
        Gets a Guild Preview Object for the guild.

        :return: guild preview object for the guild
        :rtype: discordSplash.guild.GuildPreview
        """
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.get(f'{URL}/guilds/{g_id}/preview', headers=HEADER) as r:
                preview = GuildPreview(await r.json())
                return preview

    async def update(self, name: str = None, region: str = None, verification_level: int = None,
                     default_message_notifications: int = None, explicit_content_filter: int = None,
                     afk_channel_id: int = None, afk_timeout: int = None, icon=None, owner_id: str = None,
                     splash=None, banner=None, system_channel_id: int = None, rules_channel_id: int = None,
                     public_updates_channel_id: int = None, preferred_locale: str = None):
        """
        Edits the guild.

        .. Warning::
            Some fields here will not work, depending on the guild. For example:

            - ``owner_id`` only works if the bot is the guild owner
            - ``splash`` and ``banner`` only work if the guild has a level 3 boost or is verified/partnered
            - ``system_channel`` can be disabled

        :param str name: new name of the guild
        :param str region: new voice region of the guild
        :param int verification_level: required verification level for the guild
        :param int default_message_notifications: new message notifications level for the guild
        :param int explicit_content_filter: new explicit content filter for the guild.
        :param int afk_channel_id: new channel ID for the AFK channel
        :param int afk_timeout: new number of seconds a member can be AFK before being moved into an AFK channel
        :param icon: new icon of the guild
        :param str owner_id: id of the new guild owner (bot must be guild owner for this to work.)
        :param splash: new splash image of the guild
        :param banner: new guild banner
        :param int system_channel_id: Channel ID for the system messages channel ID.
        :param int rules_channel_id: channel ids for rules.
        :param int public_updates_channel_id: new channel for moderator info (Discord community updates.)
        :param str preferred_locale: Where the guild is hosted.

        :return: Updated Guild Object
        :rtype: discordSplash.guild.Guild
        """
        json = {"name": name, "region": region, "verification_level": verification_level,
                "default_message_notifications": default_message_notifications,
                "explicit_content_filter": explicit_content_filter, "afk_channel_id": afk_channel_id,
                "afk_timeout": afk_timeout, "icon": icon, "owner_id": owner_id, "splash": splash, "banner": banner,
                "system_channel_id": system_channel_id, "rules_channel_id": rules_channel_id,
                "public_updates_channel_id": public_updates_channel_id, "preferred_locale": preferred_locale}
        for x in json:
            if json[x] is None:
                json.pop(x)

        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.patch(f'{URL}/guilds/{g_id}', json=json, headers=HEADER) as r:
                guild_new = Guild(await r.json())
                return guild_new

    async def delete(self):
        """
        deletes the guild

        bot must be guild owner

        .. Danger
            This Permanently deletes the guild.
        """
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.delete(f'{URL}/guilds/{g_id}', headers=HEADER) as r:
                guild_new = Guild(await r.json())
                return guild_new

    async def get_all_channels(self):
        """
        gets all channels in a guild

        :return: list of guild channel objects
        :rtype: [discordSplash.channel.Channel]
        """
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.get(f'{URL}/guilds/{g_id}/channels', headers=HEADER) as r:
                channels = []
                for channel_ in await r.json():
                    channels.append(channel.Channel(channel_))

                return channels

    async def create_channel(self, name: str, type: int, topic: str = None, bitrate: int = None, user_limit: int = None,
                             rate_limit_per_user: int = 0, position: int = None, permission_overwrites=None,
                             parent_id: int = None, nsfw: bool = None):
        """
        Creates a channel in the guild.

        .. Tip::
            Some parameters here are only for voice/text channels.

            Only Text:

            - ``topic``
            - ``rate_limit_per_user``

            Only Voice:

            - ``bitrate``
            - ``user_limit``

        :param str name: name of the channel
        :param int type: type of the channel
        :param str topic: Channel topic/description
        :param int bitrate: bitrate (in bits) of the channel
        :param int user_limit: limit for number of users that can join the voice channel
        :param str rate_limit_per_user: cooldown (slowmode) between sending messages
        :param permission_overwrites: permission overwrites for the channel TODO: Implement permission overwrites - implement the permission overwrites for creating channels | this will be a bit more complicated than some other feaatures | find out how this works
        :param int position: channel order position number
        :param int parent_id: id of the parent channel category
        :param bool nsfw: whether the channel is nsfw or not
        """
        json = {"name": name, "type": type, "bitrate": bitrate,
                "user_limit": user_limit,
                "rate_limit_per_user": rate_limit_per_user, "permission_overwrites": permission_overwrites,
                "position": position, "parent_id": parent_id, "nsfw": nsfw}
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.post(f'{URL}/guilds/{g_id}/channels', json=json, headers=HEADER) as r:
                return channel.Channel(await r.json())

    async def modify_channel_positions(self, data_list: list):
        """
        Adjusts Channel Positions for a guild

        .. Important::
            TODO: improve the guild.modify_channel_positions parameter - make the ``data_list`` parameter better | make it a  python object

        .. Hint::
            the ``data_list`` objects should look something like this:

            .. code-block:: json

                [
                    {
                        "id": 2349892848243,
                        "position": 1
                    },
                    {
                        "id": 7182933290482,
                        "position": 0
                    }
                ]


        :param list data_list: list of json objects for adjusting channel positions. Must be longer than 2.
        :raises: discordSplash.guild.ListLengthError if the ``data_list`` parameter is too short.
        """
        if len(data_list) < 2:
            raise ListLengthError(
                "`data_list` for coroutine discordSplash.guild.modify_channel_positions requires a length of two or more.")
        else:
            g_id = self.json['id']
            async with aiohttp.ClientSession() as cs:
                async with cs.patch(f'{URL}/guilds/{g_id}/channels', json=data_list, headers=HEADER) as r:
                    pass

    async def get_member(self, id_):
        """
        gets a Member from their id

        .. SeeAlso::
            discordSplash.member.Member

            discordSplash.guild.get_all_members()


        :param int id_: ID of the member you want to fetch
        :return: Guild member of the id
        :rtype: discordSplash.member.Member
        """

        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.get(f'{URL}/guilds/{g_id}/members/{id_}', headers=HEADER) as r:
                member_ = member.Member(await r.json())
                return member_

    async def get_members(self):
        """
        gets all members in a guild

        .. Warning::
            This may be part of privileged intents in the future.

        :return: list of all members found in the guild
        :rtype: [discordSplash.member.Member]
        """
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.get(f'{URL}/guilds/{g_id}/members', headers=HEADER) as r:
                members_ = []
                for member_ in await r.json():
                    members_.append(member.Member(member_))
                return members_

    async def modify_member(self, id_, nick: str = None, roles: list = None, mute: bool = None, deaf: bool = None,
                            channel_id: int = None):
        """
        modifies a member of the guild

        .. Hint::
            All of these fields require permission.

        .. Warning::
            If ``channel_id`` is not set, the user will be disconnected from voice.

        :param int id_: id of the member to modify
        :param str nick: new nickname of the member
        :param [list] roles: list of roles the member has
        :param bool mute: whether or not the user is muted
        :param bool deaf: whether or not the user is deafened
        :param str channel_id: id of the voice channel to move the member.

        :return: updated Member object
        :rtype: discordSplash.member.Member
        """
        json = {"nick": nick, "roles": roles, "mute": mute, "deaf": deaf, "channel_id": channel_id}
        g_id = self.json['id']
        async with aiohttp.ClientSession() as cs:
            async with cs.patch(f'{URL}/guilds/{g_id}/members/{id_}', json=json, headers=HEADER) as r:
                member_ = member.Member(r.json)
                return member_


class GuildPreview:
    """
    Guild Preview Object
    """

    def __init__(self, json):
        self.json = json

    @property
    def id(self):
        """
        ID of the guild

        :return: guild id
        :rtype: int
        """
        try:
            return self.json['id']
        except KeyError:
            return None

    @property
    def name(self):
        """
        Guild Name.

        :return: Name of the Guild
        :rtype: str
        """
        try:
            return self.json['name']
        except KeyError:
            return None

    @property
    def icon(self):
        """
        Icon hash of the guild.

        :return: guild icon hash
        :rtype: str
        """
        try:
            return self.json['icon']
        except KeyError:
            return None

    @property
    def splash(self):
        """
        Splash of the guild

        :return: guild splash hash
        :rtype: str
        """
        try:
            return self.json['splash']
        except KeyError:
            return None

    @property
    def discovery_splash(self):
        """
        discovery splash image of the guild

        :return: discovery splash hash
        :rtype: str
        """
        try:
            return self.json['discovery_splash']
        except KeyError:
            return None

    @property
    def emojis(self):
        """
        Emojis found in the Guild

        .. Important::
            TODO: Implement the discordSplash.emoji.Emoji object - add an object to represent discord Emojis. | stop sending json for emojis.

        :return: list of discordSplash.emoji.Emoji objects
        :rtype: [discordSplash.emoji.Emoji]
        """
        try:
            return self.json['emojis']
        except KeyError:
            return

    @property
    def features(self):
        """
        Guild Features

        :return: list of guild feature strings
        :rtype: [str]
        """
        try:
            return self.json['features']
        except KeyError:
            return None

    @property
    def approximate_member_count(self):
        """
        Amount of members in the guild.

        :return: Approximate number of members found in the guild
        :rtype: int
        """
        try:
            return int(self.json['approximate_member_count'])
        except KeyError:
            return None

    @property
    def approximate_presence_count(self):
        """
        Amount of online members in the guild.

        :return: Approximate number of online users in the guild
        :rtype: int
        """
        try:
            return int(self.json['approximate_presence_count'])
        except KeyError:
            return None

    @property
    def description(self):
        """
        Guild description

        :return: description of the guild
        :rtype: str
        """
        try:
            return self.json['description']
        except KeyError:
            return None


async def create(name: str, region=None, icon=None, verification_level: int = None,
                 default_message_notifications: int = None,
                 explicit_content_filter: int = None, roles=None, channels=None, afk_channel_id: int = None,
                 afk_timeout: int = None,
                 system_channel_id=None):
    """
    Creates a guild from the given parameters.

    .. SeeAlso::
        many parameters here are also explained in the discordSplash.guild.Guild

    .. Important::
        TODO: make some guild creation fields enumerators - make these fields Enums: | ``verification_level`` | ``default_message_notifications`` | ``explicit_content_filter``
        TODO: add some more guild objects - add the ``channel`` and ``role`` objects
        TODO: create_guild exception - add an exception if bot is in over 10 guilds

    .. Warning::
        this can only be used by bots in less than 10 guilds.

    :param str name: name of the guild
    :param str region: guild region
    :param str icon: icon of the guild
    :param int verification_level: verification level required to participate in the guild
    :param int default_message_notifications: when should users be pinged in the server
    :param int explicit_content_filter: adult content filter for the guild
    :param json roles: role objects for the server
    :param json channels: channel objects for the server
    :param int afk_channel_id: afk channel for the guild
    :param int afk_timeout: seconds until an AFK user is moved to the AFK channel.
    :param int system_channel_id: system channel id for messages such as boosting and users joining
    """
    json = {'name': name}

    if region is not None:
        json['region'] = region

    if icon is not None:
        json['icon'] = icon

    if roles is not None:
        json['roles'] = roles

    if verification_level is not None:
        json['verification_level'] = verification_level

    if default_message_notifications is None:
        json['default_message_notifications'] = default_message_notifications

    if explicit_content_filter is not None:
        json['explicit_content_filter'] = explicit_content_filter

    if channels is not None:
        json['channels'] = channels

    if afk_channel_id is not None:
        json['afk_channel_id'] = afk_channel_id

    if afk_timeout is not None:
        json['afk_timeout'] = afk_timeout

    if system_channel_id is not None:
        json['system_channel_id'] = system_channel_id

    async with aiohttp.ClientSession() as session:
        async with session.post(f'{URL}/guilds', headers=HEADER, json=json) as r:
            guild = Guild(await r.json())
            return guild


async def get(g_id: int, with_counts: bool = False):
    """
    Get a guild from the Guild's id.

    .. Important::
        TODO: add support for ``with_counts`` in the discordSplash.guild.Guild object - allow users to choose an amount of stuff to get

    .. Tip::

        .. SeeAlso::
            discordSplash.guild.get_preview()

        If you dont need the discordSplash.guild.Guild object, use the coroutine discordSplash.guild.get_preview.


    :param int g_id: ID of the guild you want to fetch
    :param bool with_counts: If ``True``, ``approximate_presence_count`` and ``approximate_member_count`` will also be returned
    :return: guild object
    :rtype: discordSplash.guild.Guild
    """
    async with aiohttp.ClientSession() as cs:
        async with cs.get(f'{URL}/guilds/{g_id}?with_counts={str(with_counts).lower()}', headers=HEADER) as r:
            guild = Guild(await r.json())
            return guild


async def get_guild_preview(g_id: int):
    """
    Gets a Guild Preview Object.

    :param int g_id: id of the guild
    :return: guild preview object for the guild
    :rtype: discordSplash.guild.GuildPreview
    """
    async with aiohttp.ClientSession() as cs:
        async with cs.get(f'{URL}/guilds/{g_id}/preview', headers=HEADER) as r:
            preview = GuildPreview(await r.json())
            return preview


class ListLengthError(Exception):
    pass



import discordSplash


class AuditLog:
    def __init__(self, jsondata):
        self.jsondata = jsondata

    @property
    def webhooks(self):
        """List of Webhooks found in the Audit Log
        :rtype: list
        :return: list of Webhook objects

        .. Warning::
            You will receive a list of JSON data currently.
            """
        return self.jsondata['webhooks']

    @property
    def users(self):
        """Returns a list of all users found in the Audit log
        :return: list of discordSplash.member.Member objects
        :rtype: list"""
        listUsers = []
        for user in self.jsondata['users']:
            listUsers.append(discordSplash.member.Member(user))
        return listUsers

    @property
    def integrations(self):
        """
        Returns a list of all integrations found in the Audit Log.

        .. Warning::
            May change from ``PartialIntegration`` to ``Integration``. See ``TODO`` in class discordSplash.audit_log.PartialIntegration.

        :return: list of discordSplash.audit_log.PartialIntegration objects.
        :rtype: list
        """
        listIntegrations = []
        for integration in self.jsondata['integrations']:
            listIntegrations.append(PartialIntegration(integration))
        return listIntegrations


class Entry:
    def __init__(self, jsondata):
        self.jsondata = jsondata


class PartialIntegration:
    def __init__(self, jsondata):
        """Partial integration object. Used mainly in Audit Logs

        .. Important::
            TODO: make PartialIntegration a full integration object - ### make discordSplaash.audit_log.PartialIntegration a full integration object | this will add more features on the integration object for guilds

        .. Warning::
            May move from discordSplash.audit_log.PartialIntegration to discordsplash.guild.PartialIntegration
"""
        self.jsondata = jsondata

    @property
    def id(self):
        """id of the integration
        :rtype: int"""
        return str(self.jsondata("id"))

    @property
    def name(self):
        """Name of the integration
        :rtype: str"""
        return self.jsondata("name")

    @property
    def type(self):
        """Type of the integration.
        :rtype: str"""
        return self.jsondata("type")

    @property
    def account(self):
        """integration account
        :return: Account of the integration
        :rtype: discordSplash.audit_log.Account"""
        return Account(self.jsondata("account"))


class Account:
    """
    Discord Account Object
    """

    def __init__(self, jsondata):
        self.jsondata = jsondata

    @property
    def name(self):
        """Name of the account.
        :rtype: str"""
        return self.jsondata("name")

    @property
    def id(self):
        """id of the account.
        :rtype: int"""
        return int(self.jsondata("id"))




class Channel:
    """
    Represents a Discord Channel Object.
    """
    def __init__(self, json):
        self.json = json

    @property
    def id(self):
        """
        x
        :return:
        :rtype:
        """
        try:
            return self.json['x']
        except KeyError:
            return None



import asyncio
import json
import aiohttp
import sys
from . import cfg
from enum import Enum

import aiohttp
import websockets
import traceback

PACKAGE_PARENT = '.'
try:
    from . import cfg, member, guild
    from . import opcodes as op
except (ImportError, ModuleNotFoundError):
    import cfg
    import member
    import guild
    import opcodes as op

commands = {}


class PresenceType(Enum):
    """
    Enumerator for discord PresenceTypes

    Used in the ``presenceType`` parameter for discordSplash.Presence
    """
    Game = 0
    Streaming = 1
    Listening = 2
    Custom = 4
    Competing = 5


class Presence:
    """Presence data used when connecting to Gateway

    :param PresenceType presenceType: type of presence to use.
    :param str text: Text of status to use.

    .. Hint::
        ``x = Presence('a game', discordSplash.PresenceType.Game)``

    .. Warning::
        Streaming URL's currently do not work

        Custom emojis have not been implemented in this API wrapper

        """

    def __init__(self, text: str, presenceType: PresenceType = PresenceType.Game):
        self.type_ = presenceType
        self.text_ = text

    @property
    def type(self):
        """Returns the type of the activity. Used internally

        :return: Integer from 1-5. See class discordSplash.Presence** for more info
        :rtype: int
        """

        return self.type_.value

    @property
    def text(self):
        """Activity Text. Used Internally.

        :return: The text used in the presence.
        :rtype: str
            """
        return self.text_


class InvalidTypeException(Exception):
    """Called when your response is an invalid type.
    """
    pass


class ReactionResponse:
    """
    Base class for responding to an interaction.

    .. |ss| raw:: html

        <strike>

    .. |se| raw:: html

        </strike>

    :param str content: Content of the message. Must not be more than 2000 characters.
    :param bool isEphemeral: Whether or not the message should be ephemeral (only seen by the user who created the interaction)
    :param int responseType: discord InteractionResponseType (1-5), 2 and 3 are depreciated.

    .. Note::
        [Discord InteractionResponseType](https://discord.com/developers/docs/interactions/slash-commands#interaction-response-interactionresponsetype)

        1	ACK a Ping

        [ss]2	ACK a command without sending a message, eating the user's input[se] **DEPRECIATED**

        [ss]3	respond with a message, eating the user's input[se] **DEPRECIATED**

        4	respond to an interaction with a message

        5	ACK a command without sending a message, showing the user's input


        The user's input will be shown on all of the above types unless you set ``isEphemeral`` to ``True``.


    .. Important::
        TODO: add an enumerator to ReactionResponse - Make ``responseType`` an Enumerator


        """

    def __init__(self, content: str, isEphemeral: bool = False, responseType: int = 4):
        if responseType not in [1, 2, 3, 4, 5]:
            raise InvalidTypeException(
                f"responseType {responseType} is invalid. See https://discord.com/developers/docs/interactions/slash-commands#interaction-response-interactionresponsetype for more info.")
        self.jsonContent = {
            "type": responseType,
            "data": {
                "content": str(content)
            }
        }
        if isEphemeral:
            self.jsonContent['data']['flags'] = 64

    @property
    def json(self):
        """
        Mainly used internally
        :return: JSON content for the reaction
        :rtype: json
        """
        return self.jsonContent


class ReactionData:
    """
    Data passed into all command functions

    .. Important::
        TODO: improve ReactionData - make the choices/parameters better.

    """

    def __init__(self, jsonData):
        self.jsonData = jsonData['d']

    @property
    def guild_id(self):
        """:return: the guild id
        :rtype: str

        .. Warning::
            Guilds are not implemented yet.
        """
        return int(self.jsonData["guild_id"])

    @property
    async def guild(self):
        """
        :return: Guild object that the slashcommand was executed in
        :rtype: discordSplash.guild.Guild
        """
        guildtoreturn = await guild.get(self.jsonData["guild_id"])
        f_guild = guild.Guild(guildtoreturn)
        return f_guild


    @property
    def id(self):
        """:return: the reaction id
        :rtype: str
        """
        return self.jsonData['id']

    @property
    def token(self):
        """:return: the reaction token
        :rtype: str
        .. Caution::
            - Be sure **not to send this in chat. Anyone with the token can send messages as your bot**

            - Expires after 15 minutes
        """
        return self.jsonData['token']

    @property
    def type(self):
        """:return: the reaction type
        :rtype: int

        .. Note ::
            Used for future proofing
            """
        return int(self.jsonData['type'])

    @property
    def member(self):
        """
        :return: a discordSplash.member.Member** object.
        :rtype: discordSplash.member.Member
        """
        return member.Member(self.jsonData['member'])

    @property
    def options(self):
        """
        :return: the choices/parameters for the SlashCommands.
        :rtype: list

        .. Caution::
            Currently returns a list of options. **Is not parsed yet**

        """
        options_ = []

        try:
            for x in self.jsonData['data']['options']:
                options_.append(InteractionOption(x))
        except KeyError:
            return None

    def json(self):
        """:return: the JSON. Can be used for a custom parser.
        :rtype: json"""
        return self.jsonData

    async def respond(self, data: ReactionResponse):
        """Responds to the interaction.

        Parameters:

        :param discordSplash.ReactionResponse data: Reaction Response Data


        .. Tip::
            This must be called within 3 seconds of receiving the response.

            .. Tip::
                If you do not want to immediately send a message, call this with reactionResponse ResponseType ``1``

            """
        async with aiohttp.ClientSession() as session:
            print("jsondata", self.jsonData)
            async with session.post(
                    f'https://discord.com/api/v8/interactions/{self.jsonData["id"]}/{self.jsonData["token"]}/callback',
                    json=data.json) as resp:
                pass

    async def edit(self, content: ReactionResponse):
        """
        Edits the original reaction response sent.
        :param discordSplash.ReactionResponse content: New content of the reaction response.
        """
        async with aiohttp.ClientSession as session:
            async with session.patch(
                    f'https://discord.com/api/v8/webhooks/{cfg.CLIENT_ID}/{self.jsonData["token"]}/@original',
                    json=content.json) as r:
                pass

    async def send_followup_message(self, data: ReactionResponse):
        """
        send a followup message for that interaction

        :param discordSplash.ReactionResponse data: Content of the followup message.

        .. Important::
            this needs testing.

            - To Test:
                - Ephemeral Messages
        """
        async with aiohttp.ClientSession as session:
            async with session.post(f'https://discord.com/api/v8/webhooks/{cfg.CLIENT_ID}/{self.jsonData["token"]}/',
                                    json=data.json) as r:
                pass

    async def delete_original_response(self):
        """
        delete the original reaction
        """
        async with aiohttp.ClientSession as session:
            async with session.delete(
                    f'https://discord.com/api/v8/webhooks/{cfg.CLIENT_ID}/{self.jsonData["token"]}/@original'):
                pass
    #  TODO: make it possible to edit any message from an interaction - currently it is possible to delete or edit the original response, but not any of the other responses |


class Run:
    """Runs the bot using the token

    :param str token: token of the bot to run
    :param splash.Presence presence: Presence of the bot, activated upon connection to gateway

    :raises: discordSplash.UnregisteredCommandException

    .. Hint::
        ``Run('TOKEN', Presence(text='testing', presenceType=discordSplash.PresenceType.Game))``

    .. Caution::
        Most of the methods here are only used internally.


    .. Danger::
        **Do not share your token with anyone.** If you want to collaborate on a discord bot, use a development team.
    """

    def __init__(self, token: str, presence: Presence = None):
        self.interval = None
        self.sequence = None
        self.session_id = None

        self.TOKEN = token
        cfg.TOKEN = token
        cfg.AUTH_HEADER = {"Authorization": f"Bot {token}"}
        TOKEN = self.TOKEN
        print('header 1', cfg.AUTH_HEADER)

        self.auth = {
            "token": self.TOKEN,
            "properties": {
                "$os": "Python",
                "$browser": "discordSplash",
                "$device": "discordSplash"
            },

        }
        if not presence:
            self.auth['presence'] = {
                "status": "online",
                "afk": False
            }
        else:
            print('has presence ', presence.type, presence.text)

            self.auth['presence'] = {
                "activities": [{
                    "name": presence.text,
                    "type": presence.type
                }],
                "since": 91879201,
                "status": "online",
                "afk": False
            }

        try:
            asyncio.run(self.main(False))
        except websockets.exceptions.ConnectionClosedError:
            while True:
                try:
                    asyncio.run(self.main(True))
                except websockets.exceptions.ConnectionClosedError:
                    pass
                except websockets.exceptions.ConnectionClosedOK:
                    pass
        # asyncio.get_event_loop().run_until_complete(self.hello())
        # print(self.opcode(1, self.sequence))

    async def main(self, resume=False):
        async with websockets.connect(
                'wss://gateway.discord.gg/?v=6&encoding=json') \
                as self.websocket:
            if resume is False:
                await self.hello()
                if self.interval is None:
                    print("Hello failed, exiting")
                    return
                await asyncio.gather(self.heartbeat(), self.receive())
            if resume is True:
                await self.resume()
                print('RESUMING--------------------------------')
                await asyncio.gather(self.heartbeat(), self.receive())

            # while self.interval is not None:
            #     pass

    async def receive(self):
        print("Entering receive")
        async for message in self.websocket:
            print("<", message)
            data = json.loads(message)
            if data["op"] == op.DISPATCH:
                self.sequence = int(data["s"])
                event_type = data["t"]
                if event_type == "READY":
                    print('ready')
                    cfg.CLIENT_ID = data['d']['user']['id']
                    self.session_id = data['d']['session_id']
                elif event_type == "INTERACTION_CREATE":
                    event_name = data['d']['data']['name']
                    try:
                        function = commands[event_name]
                        await function(ReactionData(data))
                    except KeyError:
                        try:
                            raise UnregisteredCommandException(
                                'One or more commands on discord are not represented on this api')
                        except UnregisteredCommandException:
                            traceback.print_exc()

    async def send(self, opcode, payload):
        data = self.opcode(opcode, payload)
        print(">", data)
        await self.websocket.send(data)

    async def heartbeat(self):
        print("Entering heartbeat")
        while self.interval is not None:
            print("Sending a heartbeat")
            await self.send(op.HEARTBEAT, self.sequence)
            await asyncio.sleep(self.interval)

    async def hello(self):
        await self.send(op.IDENTIFY, self.auth)
        print(f"hello > auth")

        ret = await self.websocket.recv()
        print(f"hello < {ret}")

        data = json.loads(ret)
        opcode = data["op"]
        if opcode != 10:
            print("Unexpected reply")
            print(ret)
            return
        self.interval = (data["d"]["heartbeat_interval"] - 2000) / 1000
        # self.interval = 5
        print("interval:", self.interval)

    def opcode(self, opcode: int, payload) -> str:
        data = {
            "op": opcode,
            "d": payload
        }
        return json.dumps(data)

    async def resume(self):
        resume_pkt = await self.create_resume_packet()
        await self.send(op.RESUME, resume_pkt)

    async def create_resume_packet(self):
        resume_blk = {
            "token": self.TOKEN,
            "session_id": self.session_id,
            "seq": self.sequence
        }
        return resume_blk


def command(name: str):
    """A decorator that is used to register a command.

    :param str name: name of the command

    .. SeeAlso::
        See ``examples`` directory on GitHub for info on usage

    .. Hint::
        Example Code:

        .. code:: python

            @discordSplash.command(name="say-hello")
            async def say_hello(data):
                await data.respond('hi')


    """

    def decorator(func):
        commands[name] = func
        return func

    return decorator


class InteractionOption:
    """
    Represents the options ('parameters') sent for the command

    .. Tip::
        You **should** check if this is a parameter using InteractionOptions.is_not_subcommand
    """

    def __init__(self, json_):
        self.json = json_

    @property
    def name(self):
        """
        Name of the interaction option
        :return: name of the interaction parameter or subcommand
        :rtype: str
        """
        return self.json['name']

    @property
    def value(self):
        """
        Value of the parameter
        :return: None if it is a subcommand group, or the value of the parameter
        :rtype: Union[int,str,None]
        """
        try:
            return self.json['value']
        except KeyError:
            return None

    @property
    def is_not_subcommand(self):
        """
        Is the option a parameter (not a subcommand)?
        :return: True if this is a parameter, false if this is a subcommand or subcommand group.
        :rtype: bool
        """
        if not self.json['options']:
            return True
        else:
            return False

    @property
    def options(self):
        """
        list of options if this option is a subcommand or subcommand group.
        :return: array of discordSplash.main.InteractionOption (this class)
        :rtype: [discordSplash.main.InteractionOption]
        """
        options__ = []
        try:
            for option in self.json['options']:
                options__.append(InteractionOption(option))
        except KeyError:
            return None


class UnregisteredCommandException(Exception):
    """
    Raised when a command is registered on the Discord API but not on discordSplash.

    .. Important::
        TODO:

        - make it a warning
    """
    pass



"""
   submodule for making requests to the API.

   Alternatively you could use something like this:

   .. code-block:: python

       async def main():
           async with aiohttp.ClientSession() as session:
               async with session.get('URL', headers=cfg.AUTH_HEADER, json={"your": "JSON", "object": "here"}) as response:
               // Parse the response

   but the methods in this class handle rate limits

   .. Caution::
       as per https://discord.com/developers/docs/topics/rate-limits#invalid-request-limit, it is **strongly recommended** that you handle ratelimits.

       I advise you use these methods rather than aiohttp as these automatically parse ratelimits.


"""

import asyncio
import time
import aiohttp
from .cfg import AUTH_HEADER as HEADER

routes = {}
guild_ids = {}
channel_ids = {}


async def ratelimit_sleeper(route: str, channel_id: int = None, guild_id: int = None):
    """
    Sleep the amount of time remaining on that route/guild/channel

    Mainly used internally.
    """
    if not channel_id:
        if not guild_id:
            if routes[route]['remaining'] == 0:
                pass
            else:
                await asyncio.sleep(routes[route]['epoch'] - int(time.time()))

        elif not guild_id in guild_ids:
            pass
        else:
            if guild_ids[guild_id]['remaining'] == 0:
                pass
            else:
                await asyncio.sleep(guild_ids[guild_id]['epoch'] - int(time.time()))
    elif not channel_id in channel_ids:
        pass
    else:
        if channel_ids[channel_id]['remaining'] == 0:
            pass
        else:
            await asyncio.sleep(channel_ids[channel_id]['epoch'] - int(time.time()))


async def ratelimit_cleanup(epoch, remaining, channel_id, guild_id, route):
    """
    Used to set the JSON dicts for the channel/guild/http route

    .. Caution::
        Tampering with this *could* lead to ``429`` responses

    This will always be overridden on the next HTTP request
    """
    if not channel_id:
        if not guild_id:
            routes[route]['remaining'] = int(remaining)
            routes[route]['epoch'] = int(epoch)
        else:
            guild_ids[guild_id]['remaining'] = int(remaining)
            guild_ids[guild_id]['epoch'] = int(epoch)
    else:
        channel_ids[channel_id]['remaining'] = int(remaining)
        channel_ids[channel_id]['epoch'] = int(epoch)


# all json formats above take into account the requests_remaining as "remaining" and epoch reset seconds as "epoch"


async def get(route: str, channel_id: int = None, guild_id: int = None):
    """
    Makes a HTTP ``GET`` request to discord's api

    :param str route: route to make the request to
    :param int channel_id: id of the channel you want to modify
    :param int guild_id: id of the guild you want to modify

    .. Tip::
        use either ``channel_id`` or ``guild_id``. **Not Both**

    """
    await ratelimit_sleeper(route, channel_id, guild_id)
    async with aiohttp.ClientSession() as session:
        async with session.get(route, headers=HEADER) as response:
            head = response.headers
            ech = head['X-RateLimit-Reset']
            rm = head['X-RateLimit-Remaining']
            await asyncio.create_task(
                ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
            return await response.json()


async def post(route, channel_id=None, guild_id=None, json=None):
    """
    Makes a HTTP ``POST`` request to discord's api

    :param str route: route to make the request to
    :param int channel_id: id of the channel you want to modify
    :param int guild_id: id of the guild you want to modify
    :param json json: json of the request body

    .. Tip::
        use either ``channel_id`` or ``guild_id``. **Not Both**

    """
    await ratelimit_sleeper(route, channel_id, guild_id)
    if not json:
        async with aiohttp.ClientSession() as session:
            async with session.post(route, headers=HEADER) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()
    else:
        async with aiohttp.ClientSession() as session:
            async with session.post(route, headers=HEADER, json=json) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()


async def patch(route, channel_id=None, guild_id=None, json=None):
    """
    Makes a HTTP ``PATCH`` request to discord's api

    :param str route: route to make the request to
    :param int channel_id: id of the channel you want to modify
    :param int guild_id: id of the guild you want to modify
    :param json json: json of the request body

    .. Tip::
        use either ``channel_id`` or ``guild_id``. **Not Both**

    """
    await ratelimit_sleeper(route, channel_id, guild_id)
    if not json:
        async with aiohttp.ClientSession() as session:
            async with session.patch(route, headers=HEADER) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()
    else:
        async with aiohttp.ClientSession() as session:
            async with session.patch(route, headers=HEADER, json=json) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()


async def delete(route, channel_id=None, guild_id=None):
    """
    Makes a HTTP ``DELETE`` request to discord's api

    :param str route: route to make the request to
    :param int channel_id: id of the channel you want to modify
    :param int guild_id: id of the guild you want to modify

    .. Tip::
        use either ``channel_id`` or ``guild_id``. **Not Both**

    """
    await ratelimit_sleeper(route, channel_id, guild_id)

    async with aiohttp.ClientSession() as session:
        async with session.delete(route, headers=HEADER) as response:
            head = response.headers
            ech = head['X-RateLimit-Reset']
            rm = head['X-RateLimit-Remaining']
            await asyncio.create_task(
                ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
            return await response.json()


async def put(route, channel_id=None, guild_id=None, json=None):
    """
    Makes a HTTP ``PUT`` request to discord's api

    :param str route: route to make the request to
    :param int channel_id: id of the channel you want to modify
    :param int guild_id: id of the guild you want to modify
    :param json json: json of the request body

    .. Tip::
        use either ``channel_id`` or ``guild_id``. **Not Both**

    """
    await ratelimit_sleeper(route, channel_id, guild_id)
    if not json:
        async with aiohttp.ClientSession() as session:
            async with session.patch(route, headers=HEADER) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()
    else:
        async with aiohttp.ClientSession() as session:
            async with session.patch(route, headers=HEADER, json=json) as response:
                head = response.headers
                ech = head['X-RateLimit-Reset']
                rm = head['X-RateLimit-Remaining']
                await asyncio.create_task(
                    ratelimit_cleanup(epoch=ech, remaining=rm, channel_id=channel_id, route=route, guild_id=guild_id))
                return await response.json()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="discordSplash",
    version="0.0.9a3",
    author="Mineinjava",
    author_email="Mineinjava@gmail.com",
    description="An API wrapper for Discord's slash commands. Written in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mineinjava/discord-splash",
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.7',
)



from .main import *
from . import ratelimit
from . import *
import os
import sys
PACKAGE_PARENT = '.'
SCRIPT_DIR = os.path.dirname(os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__))))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))
__path__ = __import__('pkgutil').extend_path(__path__, __name__)
